/*
 *	LiMon Monitor (LiMon) - Network.
 *
 *	Copyright 1994 - 2000 Neil Russell.
 *	(See License)
 */

#ifndef __NET_H__
#define __NET_H__


#include "iotypes.h"
#include <board.h>

#ifdef BIG_ENDIAN
#define ntohs(x) (x)
#define htons(x) (x)
#define ntohl(x) (x)
#else
#endif

#include <sys/types.h>
int NetInit(void);

typedef u8 u8_t;
typedef u16 u16_t;
typedef u32 u32_t;

typedef u32_t		IPaddr_t;

/**********************************************************************/
/*
 *	Ethernet card control structure
 */

/*
 *	The current receive packet handler.  Called with a pointer to the
 *	application packet, and a protocol type (PORT_BOOTPC or PORT_TFTP).
 *	All other packets are dealt with without calling the handler.
 */
typedef void		rxhand_f(u8_t *, unsigned, unsigned, unsigned);

/*
 *	A timeout handler.  Called after time interval has expired.
 */
typedef void		thand_f(void);

/*
 *	Network loop start function.
 */
typedef void		start_f(void);

/*
 *	Ethernet card functions.
 */
typedef int		init_f(void);
typedef void		send_f(volatile u8_t *, int);
typedef void		rx_f(void);
typedef void		halt_f(void);

typedef struct
{
	init_f *	init;		/* Initialize the device */
	send_f *	send;		/* Send a packet */
	rx_f *		check_rx;	/* Check for received packets */
	halt_f *	halt;		/* Shutdown card */
}
		NetCard_t;

extern NetCard_t *	NetCard;	/* The card we found */

/**********************************************************************/
/*
 *	Protocol headers.
 */

/*
 *	Ethernet header
 */
typedef struct
{
	u8_t		et_dest[6];	/* Destination node */
	u8_t		et_src[6];	/* Source node */
	u16_t		et_protlen;	/* Protocol or length */
	u8_t		et_dsap;	/* 802 DSAP */
	u8_t		et_ssap;	/* 802 SSAP */
	u8_t		et_ctl;		/* 802 control */
	u8_t		et_snap1;	/* SNAP */
	u8_t		et_snap2;
	u8_t		et_snap3;
	u16_t		et_prot;	/* 802 protocol */
}
		Ethernet_t;

#define ETHER_HDR_SIZE	14	/* Ethernet header size */
#define E802_HDR_SIZE	22	/* 802 ethernet header size */
#define PROT_IP		0x0800	/* IP protocol */
#define PROT_ARP	0x0806	/* IP ARP protocol */

/*
 *	Internet Protocol (IP) header.
 */
typedef struct
{
	u8_t		ip_hl_v;	/* header length and version */
	u8_t		ip_tos;		/* type of service */
	u16_t		ip_len;		/* total length */
	u16_t		ip_id;		/* identification */
	u16_t		ip_off;		/* fragment offset field */
	u8_t		ip_ttl;		/* time to live */
	u8_t		ip_p;		/* protocol */
	u16_t		ip_sum;		/* checksum */
	IPaddr_t	ip_src;		/* Source IP address */
	IPaddr_t	ip_dst;		/* Destination IP address */
	u16_t		udp_src;	/* UDP source port */
	u16_t		udp_dst;	/* UDP destination port */
	u16_t		udp_len;	/* Length of UDP packet */
	u16_t		udp_xsum;	/* Checksum */
}
		IP_t;

#define IP_HDR_SIZE_NO_UDP	(sizeof (IP_t) - 8)
#define IP_HDR_SIZE		sizeof (IP_t)


/*
 *	Address Resolution Protocol (ARP) header.
 */
typedef struct
{
	u16_t		ar_hrd;		/* Format of hardware address */
#   define ARP_ETHER	    1		/* Ethernet hardware address */
	u16_t		ar_pro;         /* Format of protocol address */
	u8_t		ar_hln;		/* Length of hardware address */
	u8_t		ar_pln;		/* Length of protocol address */
	u16_t		ar_op;		/* Operation */
#   define ARPOP_REQUEST    1		/* Request to resolve address */
#   define ARPOP_REPLY      2		/* Response to previous request */
	/*
	 *	The remaining fields are variable in size,
	 *	according to the sizes above, and are defined
	 *	as appropriate for specific hardware/protocol
	 *	combinations.
	 */
	u8_t		ar_data[0];
#if 0
	u8_t		ar_sha[];	/* Sender hardware address */
	u8_t		ar_spa[];	/* Sender protocol address */
	u8_t		ar_tha[];	/* Target hardware address */
	u8_t		ar_tpa[];	/* Target protocol address */
#endif // 0
}
		ARP_t;

#define ARP_HDR_SIZE	(8+20)		/* Size assuming ethernet */


/*
 *	Maximum packet size; used to allocate packet storage.
 *	TFTP packets can be 524 bytes + IP header + ethernet header.
 *	Lets be conservative, and go for 38 * 16.  (Must also be
 *	a multiple of 32 bytes).
 */
// #define PKTSIZE		1536
#define PKTSIZE		608

/*
 *	Maximum receive ring size; that is, the number of packets
 *	we can buffer before overflow happens.  Basically, this just
 *	needs to be enough to prevent a packet being discarded while
 *	we are processing the previous one.
 */
#define RINGSZ		4
#define RINGSZ_LOG2	2

/**********************************************************************/
/*
 *	Globals.
 */

/* net.c */
extern u8_t		NetOurEther[6];		/* Our ethernet address */
extern u8_t		NetServerEther[6];	/* Boot server enet address */
extern IPaddr_t		NetOurIP;		/* Our IP addr (0 = unknown) */
extern IPaddr_t		NetServerIP;		/* Server IP addr (0 = unkn) */
extern volatile u8_t *	NetTxPacket;		/* THE transmit packet */
extern volatile u8_t *	NetRxPackets[PKTBUFSRX]; /* Receive packets */
extern volatile u8_t *	NetRxPkt;		/* Current receive packet */
extern int		NetRxPktLen;		/* Current rx packet length */
extern unsigned		NetIPID;		/* IP ID (counting) */
extern const u8_t	NetBcastAddr[6];	/* Ethernet boardcast address */

extern int		NetState;		/* Network loop state */
#define NETLOOP_CONTINUE	1
#define NETLOOP_RESTART		2
#define NETLOOP_SUCCESS		3
#define NETLOOP_FAIL		4

/* Initialize the network adapter */
extern int	NetLoop(start_f * func);

/* Shutdown adapters and cleanup */
extern void	NetStop(void);

/* Load failed.  Start again. */
extern void	NetStartAgain(void);

/* Copy ethernet address */
extern void	NetCopyEther(volatile u8_t *, u8_t *);

/* Set ethernet header */
extern void	NetSetEther(volatile u8_t *, u8_t *, uint);

/* Set IP header */
extern void	NetSetIP(volatile u8_t *, IPaddr_t, int, int, int);

/* Checksum */
extern int	NetCksumOk(u8_t *, int);	/* Return true if cksum OK */
extern uint	NetCksum(u8_t *, int);		/* Calculate the checksum */

/* Set callbacks */
extern void	NetSetHandler(rxhand_f *);	/* Set RX packet handler */
extern void	NetSetTimeout(int, thand_f *);	/* Set timeout handler */

/* Transmit "NetTxPacket" */
extern void	NetSendPacket(volatile u8_t *, int);

/* Processes a received packet */
extern void	NetReceive(volatile u8_t *, int);

/* Print an IP address on the console */
extern void	NetPrintIPaddr(IPaddr_t);

/**********************************************************************/

void do_net(int argc, char **argv, void *data);
int net_bootp(void);
int net_tftpload(const char *file);
extern void	TftpStart(void);	/* Begin TFTP get */
extern unsigned char * TftpLoadAddress;	/* Place to load the image into */

#endif /* __NET_H__ */
