/* $Id: elf.c,v 1.5 2000/11/15 17:44:55 apc Exp $ */
/* Copyright 2000  AG Electronics Ltd. */
/* This code is distributed without warranty under the GPL v2 (see COPYING) */

#include <stdio.h>
#include <iotypes.h>
#include <elf.h>
#include <board.h>
#include <net.h>
#include <bsp.h>
#include <run.h>
#include <flash.h>

u32 load_elf (const char *elf_base, unsigned upper_limit)
{
    const Elf32_Ehdr *header = (const Elf32_Ehdr *) elf_base;
    u32 entry = 0xffffffff;
    unsigned ph_count;

    if (header->e_ident[EI_MAG0] != ELFMAG0 ||
	    header->e_ident[EI_MAG1] != ELFMAG1 ||
	    header->e_ident[EI_MAG2] != ELFMAG2 ||
	    header->e_ident[EI_MAG3] != ELFMAG3)
    {
	printf ("Not an ELF file\n");
	return 0xffffffff;
    }

    if (header->e_ident[EI_CLASS] != ELFCLASS32 ||
#ifdef BIG_ENDIAN
	    header->e_ident[EI_DATA] != ELFDATA2MSB ||
#else
	    header->e_ident[EI_DATA] != ELFDATA2LSB ||
#endif
	    header->e_type != ET_EXEC ||
	    header->e_machine != ELF_THIS_MACHINE)
    {
	printf ("Wrong sort of ELF file\n");
	return 0xffffffff;
    }

    printf ("No of headers: %d, initial offset: %ld, size %d\n",
	    header->e_phnum, header->e_phoff, header->e_phentsize);
    for (ph_count = 0; ph_count < header->e_phnum; ph_count++)
    {
	const Elf32_Phdr *pheader =
	    (const Elf32_Phdr *)
	    (elf_base + ph_count * header->e_phentsize + header->e_phoff);
	printf ("[%d] Physical 0x%08lx virtual 0x%08lx offset 0x%lx size 0x%lx\n",
		ph_count, pheader->p_paddr, 
                pheader->p_vaddr, pheader->p_offset, pheader->p_filesz);
	/* Linux kernel lies about its physical address */
	if (pheader->p_filesz)
	{
	    u32 real_base;
	    if (header->e_entry >= pheader->p_vaddr &&
		    header->e_entry <= pheader->p_vaddr + pheader->p_filesz)
		entry = (header->e_entry +
			 (pheader->p_paddr -
			  pheader->p_vaddr)) & ~0xc0000000;
	    real_base = pheader->p_paddr & ~0xc0000000;
            if (real_base + pheader->p_filesz > upper_limit)
            {
                printf("ELF file overlaps BIOS data\n");
                return 0xffffffff;   
            }
            memcpy((void *)real_base, elf_base + pheader->p_offset,
                   pheader->p_filesz );
            make_coherent((void *)real_base,  pheader->p_filesz);
	}

    }
    return entry;
}

void flash_elf (const char *elf_base, u32 runtime_base, 
                gbios_flash_device *flash)
{
    const Elf32_Ehdr *header = (const Elf32_Ehdr *) elf_base;
    unsigned ph_count;
    u32 lower, upper;
    
    if (header->e_ident[EI_MAG0] != ELFMAG0 ||
	    header->e_ident[EI_MAG1] != ELFMAG1 ||
	    header->e_ident[EI_MAG2] != ELFMAG2 ||
	    header->e_ident[EI_MAG3] != ELFMAG3)
    {
	printf ("Not an ELF file\n");
	return;
    }

    if (header->e_ident[EI_CLASS] != ELFCLASS32 ||
#ifdef BIG_ENDIAN
	    header->e_ident[EI_DATA] != ELFDATA2MSB ||
#else
	    header->e_ident[EI_DATA] != ELFDATA2LSB ||
#endif
	    header->e_type != ET_EXEC ||
	    header->e_machine != ELF_THIS_MACHINE)
    {
	printf ("Wrong sort of ELF file\n");
	return;
    }

    printf ("No of headers: %d, initial offset: %ld, size %d\n",
	    header->e_phnum, header->e_phoff, header->e_phentsize);
    lower = flash->size;
    upper = 0;
    for (ph_count = 0; ph_count < header->e_phnum; ph_count++)
    {
	const Elf32_Phdr *pheader =
	    (const Elf32_Phdr *)
	    (elf_base + ph_count * header->e_phentsize + header->e_phoff);
	printf ("[%d] Physical 0x%08lx virtual 0x%08lx offset 0x%lx size 0x%lx\n",
		ph_count, pheader->p_paddr, 
                pheader->p_vaddr, pheader->p_offset, pheader->p_filesz);
	/* Linux kernel lies about its physical address */
	if (pheader->p_filesz)
	{
	    u32 real_base = pheader->p_paddr - runtime_base;
            if (real_base + pheader->p_filesz > flash->size)
            {
                printf("ELF file too large for FLASH\n");
                return;   
            }
            printf("Section in FLASH from 0x%08lx to 0x%08lx\n",
                    real_base, real_base + pheader->p_filesz);
            if (real_base < lower)
                lower = real_base;
            if (real_base + pheader->p_filesz > upper)
                upper = real_base + pheader->p_filesz;
	}
    }
        printf("Flash erase from 0x%08lx to 0x%08lx: ", lower, upper);
        if (flash->fn->erase(flash->data, lower, upper) < 0)
        {
            printf("Failed\n");
            return;
        }
        else
            printf("Complete\n");
    for (ph_count = 0; ph_count < header->e_phnum; ph_count++)
    {
	const Elf32_Phdr *pheader =
	    (const Elf32_Phdr *)
	    (elf_base + ph_count * header->e_phentsize + header->e_phoff);
	/* Linux kernel lies about its physical address */
	if (pheader->p_filesz)
	{
	    u32 real_base = pheader->p_paddr - runtime_base;
            printf("Section in FLASH from 0x%08lx to 0x%08lx: ",
                    real_base, real_base + pheader->p_filesz);
            if (flash->fn->program(flash->data, real_base, 
                     elf_base + pheader->p_offset, pheader->p_filesz) < 0)
            {
                    printf("Failed\n");
                    return;
                }
                else
                    printf("Complete\n");
	}

    }
}


void do_unpack(int argc, char **argv)
{
    executable_start_address = load_elf(TftpLoadAddress, bios_base);
}
