/*
 *	Linux Monitor (LiMon) - Networking.
 *
 *	Copyright 1994 - 2000 Neil Russell.
 *	(See License)
 */
#undef DEBUG
#include	<net.h>
#include <stdio.h>
#include <board.h>
#include <console.h>
#include <timer.h>
#include <dictionary.h>
#include <bsp.h>
#include <run.h>
#include "bootp.h"

NetCard_t *		NetCard;		/* Card control structure */
u8_t			NetOurEther[6];		/* Our ethernet address */
u8_t			NetServerEther[6];	/* Boot server enet address */
IPaddr_t		NetOurIP;		/* Our IP addr (0 = unknown) */
IPaddr_t		NetServerIP;		/* Our IP addr (0 = unknown) */
volatile u8_t *		NetRxPkt;		/* Current receive packet */
int			NetRxPktLen;		/* Current rx packet length */
unsigned		NetIPID;		/* IP packet ID */
RODATA const u8_t			NetBcastAddr[6] =	/* Ethernet bcast address */
				{ 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
int			NetState;		/* Network loop state */

volatile u8_t		PktBuf[(PKTBUFSRX+1) * PKTSIZE + PKTALIGN];
volatile u8_t *		NetTxPacket;		/* THE transmit packet */
volatile u8_t *		NetRxPackets[PKTBUFSRX]; /* Receive packets */

static rxhand_f *	packetHandler;		/* Current RX packet handler */
static thand_f *	timeHandler;		/* Current timeout handler */
static u32_t		timeValue;		/* Current timeout value */

/**********************************************************************/
/*
 *	Main network processing loop.
 */

int
NetLoop(start_f * start_func)
{
	if (!NetTxPacket)
	{
		int		i;

		/*
		 *	Setup packet buffers, aligned correctly.
		 */
		NetTxPacket = &PktBuf[0] + (PKTALIGN - 1);
		NetTxPacket -= (u32_t)NetTxPacket % PKTALIGN;
		for (i = 0; i < PKTBUFSRX; i++)
			NetRxPackets[i] = NetTxPacket + (i+1)*PKTSIZE;
	}

	/*
	 *	Make sure we have a network adapter.
	 */
	if (!NetCard && !NetInit())
		return -1;

    restart:
	NetState = NETLOOP_CONTINUE;

	/*
	 *	Start the ball rolling with the given start function.  From
	 *	here on, this code is a state machine driven by received
	 *	packets and timer events.
	 */
	(*start_func)();

	/*
	 *	Main packet reception loop.  Loop receiving packets until
	 *	someone sets `NetQuit'.
	 */
	for (;;)
	{
		/*
		 *	Check the ethernet for a new packet.  The ethernet
		 *	receive routine will process it.
		 */
		(*NetCard->check_rx)();

		/*
		 *	Check the keyboard for an ESC.  Quit if we get one.
		 */
		if (console_input() != -1)
		{
			printf("\nAbort\n");
			return 0;
		}

		/*
		 *	Check for a timeout, and run the timeout handler
		 *	if we have one.
		 */
		if (timeHandler && ticks_since_boot() > timeValue)
		{
			thand_f *	x;

			x = timeHandler;
			timeHandler = (thand_f *)0;
			(*x)();
		}

		switch (NetState)
		{
		case NETLOOP_RESTART:
			goto restart;

		case NETLOOP_SUCCESS:
			return 1;

		case NETLOOP_FAIL:
			return 0;
		}
	}
}

/**********************************************************************/

static void
startAgainTimeout(void)
{
	NetState = NETLOOP_RESTART;
}


static void
startAgainHandler(u8_t * pkt, unsigned dest, unsigned src, unsigned len)
{
	/* Totally ignore the packet */
}


void
NetStartAgain(void)
{
	NetServerIP = 0;
	NetOurIP = 0;
	NetSetTimeout(60 * get_hz(), startAgainTimeout);
	NetSetHandler(startAgainHandler);
}

/**********************************************************************/
/*
 *	Miscelaneous bits.
 */

void
NetSetHandler(rxhand_f * f)
{
	packetHandler = f;
}


void
NetSetTimeout(int iv, thand_f * f)
{
	if (iv == 0)
		timeHandler = (thand_f *)0;
	else
	{
		timeHandler = f;
		timeValue = ticks_since_boot() + iv;
	}
}


void
NetSendPacket(volatile u8_t * pkt, int len)
{
	(*NetCard->send)(pkt, len);
}


void
NetReceive(volatile u8_t * pkt, int len)
{
	Ethernet_t *	et;
	IP_t *		ip;
	ARP_t *		arp;
	int		x;


	NetRxPkt = pkt;
	NetRxPktLen = len;
	et = (Ethernet_t *)pkt;

	x = ntohs(et->et_protlen);
	if (x < 1514)
	{
		/*
		 *	Got a 802 packet.  Check the other protocol field.
		 */
		x = ntohs(et->et_prot);
		ip = (IP_t *)(pkt + E802_HDR_SIZE);
		len -= E802_HDR_SIZE;
	}
	else
	{
		ip = (IP_t *)(pkt + ETHER_HDR_SIZE);
		len -= ETHER_HDR_SIZE;
	}

#ifdef DEBUG
	printf("Receive from protocol 0x%x\n", x);
#endif

	switch (x)
	{
	case PROT_ARP:
		/*
		 *	The only type of ARP packet we deal with is a request
		 *	for our ethernet address.  We can only respond if we
		 *	know our IP address.
		 */
		arp = (ARP_t *)ip;
		if (len < ARP_HDR_SIZE)
		{
#ifdef DEBUG
		printf("bad length %d < %d\n", len, ARP_HDR_SIZE);
#endif
			return;
		}
		if (ntohs(arp->ar_op) != ARPOP_REQUEST)
			return;
		if (ntohs(arp->ar_hrd) != ARP_ETHER)
			return;
		if (ntohs(arp->ar_pro) != PROT_IP)
			return;
		if (arp->ar_hln != 6)
			return;
		if (arp->ar_pln != 4)
			return;

		if (NetOurIP == 0 ||
		    *((IPaddr_t *)&arp->ar_data[16]) != NetOurIP)
			return;

		NetSetEther((u8_t *)et, et->et_src, PROT_ARP);
		arp->ar_op = ntohs(ARPOP_REPLY);
		NetCopyEther(&arp->ar_data[10], &arp->ar_data[0]);
		NetCopyEther(&arp->ar_data[0], NetOurEther);
		*(IPaddr_t *)(&arp->ar_data[16]) =
					*(IPaddr_t *)(&arp->ar_data[6]);
		*(IPaddr_t *)(&arp->ar_data[6]) = NetOurIP;
		NetSendPacket((u8_t *)et, ((u8_t *)arp - pkt) + ARP_HDR_SIZE);
		break;

	case PROT_IP:
#ifdef DEBUG
		printf("Got IP\n");
#endif
		if (len < IP_HDR_SIZE)
			return;
		if (len < ntohs(ip->ip_len))
		{
#ifdef DEBUG
			printf("len bad %d < %d\n", len, ntohs(ip->ip_len));
#endif
			return;
		}
		len = ntohs(ip->ip_len);
#ifdef DEBUG
		printf("len=%d, v=0x%02x\n", len, ip->ip_hl_v & 0xff);
#endif
		if ((ip->ip_hl_v & 0xf0) != 0x40)
			return;
		if (ip->ip_off & ntohs(0x1fff)) /* Can't deal w/ fragments */
			return;
		if (!NetCksumOk((u8_t *)ip, IP_HDR_SIZE_NO_UDP / 2))
		{
#ifdef DEBUG
			printf("checksum bad\n");
#endif
			return;
		}
		if (NetOurIP && ip->ip_dst != NetOurIP && ip->ip_dst != 0xffffffff)
			return;
		if (ip->ip_p != 17)		/* Only UDP packets */
			return;

		/*
		 *	IP header OK.  Pass the packet to the current handler.
		 */
		(*packetHandler)((u8_t *)ip +IP_HDR_SIZE,
						ntohs(ip->udp_dst),
						ntohs(ip->udp_src),
						ntohs(ip->udp_len) - 8);
		break;
	}
}

/**********************************************************************/

int
NetCksumOk(u8_t * ptr, int len)
{
	return !((NetCksum(ptr, len) + 1) & 0xfffe);
}


unsigned
NetCksum(u8_t * ptr, int len)
{
	u32_t		xsum;

	xsum = 0;
	while (len-- > 0)
		xsum += *((u16_t *)ptr)++;
	xsum = (xsum & 0xffff) + (xsum >> 16);
	xsum = (xsum & 0xffff) + (xsum >> 16);
	return xsum & 0xffff;
}


void
NetCopyEther(volatile u8_t * to, u8_t * from)
{
	int		i;

	for (i = 0; i < 6; i++)
		*to++ = *from++;
}


void
NetSetEther(volatile u8_t * xet, u8_t * addr, uint prot)
{
	volatile Ethernet_t *	et = (Ethernet_t *)xet;

	NetCopyEther(et->et_dest, addr);
	NetCopyEther(et->et_src, NetOurEther);
	et->et_protlen = ntohs(prot);
}


void
NetSetIP(volatile u8_t * xip, IPaddr_t dest, int dport, int sport, int len)
{
	volatile IP_t *		ip = (IP_t *)xip;

	/*
	 *	If the data is an odd number of bytes, zero the
	 *	byte after the last byte so that the checksum
	 *	will work.
	 */
	if (len & 1)
		xip[IP_HDR_SIZE + len] = 0;

	/*
	 *	Construct an IP and UDP header.
			(need to set no fragment bit - XXX)
	 */
	ip->ip_hl_v = 0x45;	/* IP_HDR_SIZE / 4 (not including UDP) */
	ip->ip_tos = 0;
	ip->ip_len = ntohs(IP_HDR_SIZE + len);
	ip->ip_id = ntohs(NetIPID++);
	ip->ip_off = ntohs(0x4000);		/* No fragmentation */
	ip->ip_ttl = 255;
	ip->ip_p = 17;		/* UDP */
	ip->ip_sum = 0;
	ip->ip_src = NetOurIP;
	ip->ip_dst = dest;
	ip->udp_src = ntohs(sport);
	ip->udp_dst = ntohs(dport);
	ip->udp_len = ntohs(8 + len);
	ip->udp_xsum = 0;
	ip->ip_sum = ~NetCksum((u8_t *)ip, IP_HDR_SIZE_NO_UDP / 2);
}


void
NetPrintIPaddr(IPaddr_t x)
{
	x = ntohl(x);
	printf("%d.%d.%d.%d",	(int)((x >> 24) & 0xff),
				(int)((x >> 16) & 0xff),
				(int)((x >> 8) & 0xff),
				(int)((x >> 0) & 0xff));
}
extern void BootpRequest(void);

int net_bootp(void)
{
    return NetLoop(BootpRequest);   
}
void do_bootp(int argc, char **argv, void *data)
{
    int result = net_bootp();
    printf("Result %d\n", result);
}

int net_tftp(const char *name)
{
    int result;
    if (! NetServerIP)
        if (! net_bootp())
            return 0;
    if (name)
        memcpy(BootFile, name, sizeof(BootFile));
    code_loaded = NetLoop(TftpStart);
    return code_loaded;
}

void do_tftp(int argc, char **argv, void *data)
{
    extern void TftpStart(void);
    int result = 0;
    if (argc)
        result = net_tftp(argv[0]);
    else
        result = net_tftp(NULL);
    printf("Result %d\n", result);
 /*   if (result)
        result = load_elf(TftpLoadAddress, 0x8000000);
    printf("Elf result 0x%08x\n", result);
    if (result != 0xffffffff)
        printf("Command 0x%08x\n", *(unsigned *)result);
    bsp_run_executable(result, NULL, 0, 0); */
}

const gbios_dictionary net_dictionary[] = {
    {"bootp", "Get BOOTP information from server", do_bootp, NULL},
    {"tftp", "Load boot file from server", do_tftp, NULL},
    {NULL}
};

const gbios_helpdata net_helpdata = {
    "Network Commands\n", net_dictionary, NULL
};

void do_net(int argc, char **argv, void *data)
{
    if (argc)
        search_dictionary(net_dictionary, argc, argv);
    else
        printf("Specify net subcommand\n");    
}
