/*
 *      NE2100 cards - based on the AMD 79C960 PCnet/ISA (LANCE).
 *
 *      Copyright 1994, 1995 Neil Russell.
 *      (See License)
 */

#undef DEBUG
#include <pci.h>
#include <io.h>
#include <stdio.h>
#include <timer.h>
#include <board.h>
#include <malloc.h>

#define INB(x)		in_8(x)
#define OUTB(x, y)	out_8(x, y)
#define INW(x)		in_16(x)
#define OUTW(x, y)	out_16(x, y)

#include	"net.h"

#define ptr_to_lower(x) ((x) & 0xffff)
#define ptr_to_upper(x) ((x) >> 16)

/*
 *      NE2100 board registers
 */
#define HW_ROM		0x00	/* IEEE Address PROM */
#define HW_RDP		0x10	/* Register data port */
#define HW_RAP		0x12	/* Register/ISA address port */
#define HW_RESET	0x14	/* Reset register */
#define HW_BDP		0x16	/* ISA data register */
/*
 *      Control/status register definitions
 *
 *      CSR 0
 */
#define CSR0_INIT	0x0001	/* Initialise */
#define CSR0_STRT	0x0002	/* Start */
#define CSR0_STOP	0x0004	/* Stop */
#define CSR0_TDMD	0x0008	/* Transmit demand */
#define CSR0_TXON	0x0010	/* Transmitter on */
#define CSR0_RXON	0x0020	/* Receiver on */
#define CSR0_INEA	0x0040	/* Interrupt enable */
#define CSR0_INTR	0x0080	/* Interrupt flag */
#define CSR0_IDON	0x0100	/* Initialisation done */
#define CSR0_TINT	0x0200	/* Transmitter interrupt */
#define CSR0_RINT	0x0400	/* Receive interrupt */
#define CSR0_MERR	0x0800	/* Memory error */
#define CSR0_MISS	0x1000	/* Missed packet */
#define CSR0_CERR	0x2000	/* Collision error */
#define CSR0_BABL	0x4000	/* Babble */
#define CSR0_ERR	0x8000	/* Error */

/*
 *      CSR 3
 */
#define CSR3_BCON	0x0001	/* Byte control */
#define CSR3_ACON	0x0002	/* ALE control */
#define CSR3_BSWP	0x0004	/* Byte swap */

/*
 *      LANCE initialisation block
 */

typedef struct Init_t
{
#define  MODE_DRX	 0x0001	/* Disable receiver */
#define  MODE_DTX	 0x0002	/* Disable transmitter */
#define  MODE_LOOP	 0x0004	/* Loopback mode select */
#define  MODE_DTCR	 0x0008	/* Disable transmit CRC generation */
#define  MODE_COLL	 0x0010	/* Force collision */
#define  MODE_DRTY	 0x0020	/* Disable retry */
#define  MODE_INTL	 0x0040	/* Internal loopback */
#define  MODE_PROM	 0x8000	/* Promiscuous mode */
    u16 mode;
    u16 tlen_rlen;
    u8  paddr[6];
    u16 reserved;
    u32 filter[2];
    /* Receive and transmit ring base, along with extra bits. */    
    u32 rx_ring;
    u32 tx_ring;
}
Init_t;

/*
 *      LANCE receive ring buffer descriptor
 */
typedef struct RXbuf_t
{
#define  RX_ENP		 0x0100	/* End of packet */
#define  RX_STP		 0x0200	/* Start of packet */
#define  RX_BUFF	 0x0400	/* Buffer error */
#define  RX_CRC		 0x0800	/* CRC error */
#define  RX_OFLO	 0x1000	/* Overflow error */
#define  RX_FRAM	 0x2000	/* Framing error */
#define  RX_ERR		 0x4000	/* Error summary */
#define  RX_OWN		 0x8000	/* 0=owned by host; 1=owned by lance */
    u32 base;
    s16 buf_length;
    s16 status;    
    u32 msg_length;
    u32 reserved;
} RXbuf_t;

/*
 *      LANCE transmit ring buffer descriptor
 */
typedef struct TXbuf_t
{
#define  TX_ENP		 0x0100	/* End of packet */
#define  TX_STP		 0x0200	/* Start of packet */
#define  TX_DEF		 0x0400	/* Defered */
#define  TX_ONE		 0x0800	/* Exactly one retry was needed */
#define  TX_MORE	 0x1000	/* More than one retry was needed */
#define  TX_ERR		 0x4000	/* Error summary */
#define  TX_OWN		 0x8000	/* 0=owned by host; 1=owned by lance */
    u32 base;
    s16 length;
    s16 status;
    u32 misc;
    u32 reserved;
} TXbuf_t;


static ioaddr lance_port;	/* I/O port of the card */

/* static volatile Init_t lance_init;	*/ /* Initialisation block */
static int lance_ptr;		/* Receive ring pointer */

/*
 *      LANCE buffer descriptors must be 16 byte aligned.
 *      We will kludge it (See lance_init()).
 */

static u8 lance_xxx[(RINGSZ + 1 + 1) * sizeof (RXbuf_t)];
static volatile TXbuf_t *lance_tx;	/* Transmit buffer descriptor */
static volatile RXbuf_t *lance_rx;	/* Receive ring buffer desciptors */

/*
 *      PCI stuff.
 */
#define PCI_AMD_VENDOR_ID	0x1022	/* AMD's vendor ID */
#define PCI_LANCE_DEVICE_ID	0x2000	/* Am79C970 device ID */

/**********************************************************************/
/*
 *      Support routines
 */

/*
 *      Output 'd' to LANCE register 'r'.
 */
static void routw (int r, int d)
{
    OUTW (lance_port + HW_RAP, r);
    OUTW (lance_port + HW_RDP, d);
}


/*
 *      Input from LANCE register 'r'.
 */
static unsigned short rinw (int r)
{
    OUTW (lance_port + HW_RAP, r);
    return INW (lance_port + HW_RDP);
}

static void boutw(int r, u16 d)
{
    OUTW(lance_port + HW_RAP, r);  
    OUTW(lance_port + HW_BDP, d);
}
static unsigned short binw (int r)
{
    OUTW (lance_port + HW_RAP, r);
    return INW (lance_port + HW_BDP);
}

/**********************************************************************/
/*
 *      Probe a board.  If its not found, set a flag so that we
 *      don't try to use it.
 */
static int lance_probe (ioaddr port)
{
    int i;
    int address_blank = 1;
    volatile Init_t *lance_init = malloc(sizeof(Init_t));

    if (! lance_init)
        return 0;
    
    lance_port = port;

    /*
     *      Reset the card.
     */
    (void) INW (lance_port + HW_RESET);
#ifdef DEBUG
    printf("Reset cycle to 0x%x, stack 0x%08x\n", lance_port + HW_RESET,
            &i);
#endif
    /*
     *      Wait up to 1 second for the reset to take effect.  Check
     *      for completion every 100ms.
     */
#ifdef DEBUG
    printf ("probe waits 1 second\n");
#endif /* DEBUG */
    for (i = 0; i < 10; i++)
    {
	if (rinw (0) == CSR0_STOP)
	    break;

	/*
	 *      Sleep for 100ms.
	 */
	sleep (get_hz() / 10);
    }
    if (i == 10)
    {
#ifdef DEBUG
	printf ("probe wait expired\n");
#endif /* DEBUG */
	return 0;
    }
#ifdef DEBUG
    printf ("probe wait done\n");
#endif /* DEBUG */
    /* Set swstyle 32 bit */
    boutw(20, 0x8002);
    /* Burst mode */
    boutw(18, binw(18) | 0x60);
    /*
     *      Get the Ethernet address from the PROM.
     */
    for (i = 0; i < 6; i++)
    {
	NetOurEther[i] = INB (lance_port + HW_ROM + i);
	lance_init->paddr[i] = NetOurEther[i];
        if (NetOurEther[i] != 0xff)
            address_blank = 0;
    }
    if (address_blank)
    {
        lance_init->paddr[0] = NetOurEther[0] = 0x02;
        boutw(18, 0x0860 );
        boutw(25, 23);
        boutw(26, 12);
        boutw(27, 1);
    }    
#ifdef DEBUG
    printf ("lance 0x%lx: Ethernet address %02x:%02x:%02x:%02x:%02x:%02x\n",
	    lance_port,
	    NetOurEther[0],
	    NetOurEther[1],
	    NetOurEther[2], NetOurEther[3], NetOurEther[4], NetOurEther[5]);
#endif /* DEBUG */

    /*
     *      Setup the rest of the initialisation block.
     */
    /* lp->init.mode = Init_t::DRX;     */    /* Disable RX for testing */
    lance_init->mode = 0;

    lance_init->rx_ring = cpu_to_le32(local_to_pci(&lance_rx[0]));
    lance_init->tx_ring = cpu_to_le32(local_to_pci (lance_tx));
    lance_init->tlen_rlen = cpu_to_le16(RINGSZ_LOG2 << 4);	/* One TX entry */

#ifdef DEBUG
    printf("lance_init = %p, mem =", lance_init);
    for(i = 0; i < 0x20; i++)
        printf(" %02x", *(i + (char *)lance_init));
    printf("\n");
#endif /* DEBUG */

    /*
     *      Set the STOP bit.
     */
    routw (0, CSR0_STOP);

    /*
     *      Set the bus mode for an ISA bus.  On the original lance,
     *      and the PCI lance, this has the effect of putting the
     *      device in little endian mode as well as disabling all
     *      interrupts.  On the PCnet/ISA, it just disables interrrupts.
     */
    routw (3, 0);

    /*
     *      Check that the card is there again just in case.
     *      Enabling DMA when there is no device is enough
     *      to hang the whole system.
     */
    if (rinw (3) != 0)
	return 0;

    /*
     *      Set the initialisation buffer address.
     */
    routw (1, ptr_to_lower(local_to_pci(lance_init)));
    routw (2, ptr_to_upper(local_to_pci(lance_init)));

    /*
     *      Set the INIT bit;  LANCE reads init block.
     */
    routw (0, CSR0_INIT);

    /*
     *      Wait for up to 3 seconds for the initialisation to complete.
     *      Check for completion every 100ms.
     */
#ifdef DEBUG
    printf ("wait 3 seconds\n");
#endif /* DEBUG */
    for (i = 0; i < 30; i++)
    {
	if (rinw (0) & CSR0_IDON)
	    break;

	/*
	 *      Sleep for 100ms
	 */
	sleep (get_hz() / 10);
    }
    if (i == 30)
    {
	/*
	 *      Failed to initialise; reset the card for good measure,
	 *      disable the DMA channel (only on ISA) and exit.
	 */
	(void) INW (lance_port + HW_RESET);
#ifdef DEBUG
	printf ("3 second wait - expired\n");
#endif /* DEBUG */
	return 0;
    }
#ifdef DEBUG
    printf ("3 second wait over\n");
#endif /* DEBUG */

    /*
     *      We can now commit to using this card.  Start the LANCE.
     *      this is where we would also enable interrupts if we were
     *      using them.  Assuming interrupts were enabled, we would
     *      now immediately get an IDON interrupt.
     */
    /* routw(0, CSR_t::STRT | CSR_t::INEA); */
    routw (0, CSR0_STRT);

    printf ("Found PCnet/PCI at I/O address 0x%lx", lance_port);

    printf (" (%02x:%02x:%02x:%02x:%02x:%02x)\n",
	    NetOurEther[0], NetOurEther[1],
	    NetOurEther[2], NetOurEther[3], NetOurEther[4], NetOurEther[5]);

    /*
     *      Seems that if we transmit a packet immediately, it
     *      doesn't actually make it to the wire.  So we delay a
     *      little to let the LANCE get its act together.
     */
    sleep (get_hz() / 3);
    printf("Probe over\n");
    free(lance_init);
    return 1;
}


int ne2100_init (void)
{
    int i;
    u8_t *xp;

    /*
     *      Setup the ring buffers.
     */
    xp = &lance_xxx[0] + 15;
    xp = (u8_t *) ((unsigned) xp & ~15);
    lance_tx = (TXbuf_t *) xp;	/* Now 16 byte aligned */
    lance_rx = (RXbuf_t *) (lance_tx + 1);

    for (i = 0; i < RINGSZ; i++)
    {
	lance_rx[i].base = cpu_to_le32(local_to_pci(NetRxPackets[i]));
	lance_rx[i].buf_length = cpu_to_le16(-PKTSIZE);
	lance_rx[i].status = cpu_to_le16(RX_OWN);
    }
    lance_tx->status = 0;
    lance_ptr = 0;

#if 1
    /*
     *      Look for PCI devices.
     */
    {
	struct pci_device *dev;

	dev = pci_find_device (PCI_AMD_VENDOR_ID, PCI_LANCE_DEVICE_ID, 0);
	if (dev)
	{

	    /* Probe the PCnet/PCI chip */
	    if (lance_probe (dev->region[0]))
		return 1;
	}
    }
#endif


    return 0;
}


void ne2100_send (volatile u8_t * pkt, int len)
{
    u16 csr0 = rinw(0);
    unsigned end_time;
#ifdef DEBUG
    
    routw (0, CSR0_STRT | CSR0_CERR);
    printf ("Sending %p len %d ... ", pkt, len);
        printf("CSR0: 0x%04x\n", csr0);
#endif /* DEBUG */
    
    if (! (csr0 & 0x10))
    {
        printf("Transmit disabled\n");
        NetState = NETLOOP_FAIL;
        return;
    }
    
    if (len < 60)
	len = 60;
    lance_tx->base = cpu_to_le32(local_to_pci (pkt));
    lance_tx->length = cpu_to_le16(-len);
    lance_tx->status = cpu_to_le16(TX_STP | TX_ENP);

#ifdef DEBUG
#if 1
    {
	u8_t *p;
	int i;

	printf ("lance_tx=%p; mem =", lance_tx);
	p = (u8_t *) lance_tx;
	for (i = 0; i < 12; i++)
	    printf (" %02x", *p++);
	printf ("\n");
    }
#endif
#endif /* DEBUG */
    lance_tx->status = cpu_to_le16(TX_OWN | TX_STP | TX_ENP);
    routw(0, CSR0_TDMD);
    end_time = 2*get_hz() + ticks_since_boot();
    while (le16_to_cpu(lance_tx->status) & TX_OWN)
    {
        if (ticks_since_boot() > end_time)
        {
            NetState = NETLOOP_FAIL;
            break;
        }  
    
    }
    
#ifdef DEBUG
    if (! (le16_to_cpu(lance_tx->status) & TX_OWN))
        printf ("Sent ");
    else
        printf("Timeout ");
    {
	u8_t *p;
	int i;

	printf ("lance_tx=%p; mem =", lance_tx);
	p = (u8_t *) lance_tx;
	for (i = 0; i < 12; i++)
	    printf (" %02x", *p++);
	printf ("\n");
    }
#endif /* DEBUG */
}


void ne2100_check_rx (void)
{
    for (;;)
    {
	volatile RXbuf_t *bdp;

	bdp = &lance_rx[lance_ptr];

	/*
	 *      Check the current buffer.  If its still owned by
	 *      the LANCE, quit now.
	 */
	if (le16_to_cpu(bdp->status) & RX_OWN)
	    break;

#ifdef DEBUG
	printf
	    ("lance: received a buffer %p (mode 0x%x, mcnt=%ld, addr=0x%lx)\n",
	     bdp, le16_to_cpu(bdp->status), le32_to_cpu(bdp->msg_length), le32_to_cpu(bdp->base));
#endif /* DEBUG */

	/*
	 *      Check for errors.
	 */
	if ((le16_to_cpu(bdp->status) & (RX_ERR | RX_STP | RX_ENP)) != (RX_STP | RX_ENP))
	    goto done;

	/*
	 *      Pass the packet up to the protocol layers.
	 */
	NetReceive (NetRxPackets[lance_ptr], le32_to_cpu(bdp->msg_length) - 4);

	/*
	 *      Give the buffer back to the LANCE.
	 */
      done:
	bdp->base = cpu_to_le32(local_to_pci (NetRxPackets[lance_ptr]));
	bdp->buf_length = cpu_to_le16(-PKTSIZE);
	bdp->status = cpu_to_le16(RX_OWN);
	if (++lance_ptr >= RINGSZ)
	    lance_ptr = 0;
    }
}


void ne2100_halt (void)
{
    /*
     *      Stop LANCE and disable the DMA channel.
     */
    routw (0, CSR0_STOP);
}


const NetCard_t NetNE2100 = {
    ne2100_init,
    ne2100_send,
    ne2100_check_rx,
    ne2100_halt
};
