/*
 *	LiMon - TFTP.
 *
 *	Copyright 1994, 1995, 2000 Neil Russell.
 *	(See License)
 */

#include	<net.h>
#include    <stdio.h>
#include <console.h>
#include <elf.h>
#include <timer.h>
#include	"bootp.h"


#define WELL_KNOWN_PORT	69		/* Well known TFTP port # */
#define TIMEOUT		6		/* Seconds to timeout for a lost pkt */
#define TIMEOUT_COUNT	10		/* # of timeouts before giving up */
					/* (for checking the image size) */
#define NDOTS		65		/* Number of "loading" dots */

/*
 *	TFTP operations.
 */
#define TFTP_RRQ	1
#define TFTP_WRQ	2
#define TFTP_DATA	3
#define TFTP_ACK	4
#define TFTP_ERROR	5


unsigned char * TftpLoadAddress;	/* Place to load the image into */
static int	TftpServerPort;		/* The UDP port at their end */
static int	TftpOurPort;		/* The UDP port at our end */
static int	TftpTimeoutCount;
static unsigned	TftpBlock;
static unsigned	TftpLastBlock;
static int	TftpState;
static unsigned TftpStarted;
#define STATE_RRQ	1
#define STATE_DATA	2
#define STATE_TOO_LARGE	3
#define STATE_BAD_MAGIC	4


static void
store_block(unsigned block, u8_t * src, unsigned len)
{
	(void)memcpy((void *)(TftpLoadAddress + block * 512), src, len);
}

/**********************************************************************/

static void
TftpSend(void)
{
	volatile u8_t *	pkt;
	volatile u8_t *	xp;
	int			len;


	/*
	 *	We will always be sending some sort of packet, so
	 *	cobble together the packet headers now.
	 */
	pkt = NetTxPacket + ETHER_HDR_SIZE + IP_HDR_SIZE;

	switch (TftpState)
	{
	case STATE_RRQ:
		xp = pkt;
		*((u16_t *)pkt)++ = htons(TFTP_RRQ);
		strcpy((char *)pkt, &BootFile[0]);
		pkt += strlen(BootFile) + 1;
		strcpy((char *)pkt, "octet");
		pkt += 5 /*strlen("octet")*/ + 1;
		len = pkt - xp;
		break;

	case STATE_DATA:
		xp = pkt;
		*((u16_t *)pkt)++ = htons(TFTP_ACK);
		*((u16_t *)pkt)++ = htons(TftpBlock);
		len = pkt - xp;
		break;

	case STATE_TOO_LARGE:
		xp = pkt;
		*((u16_t *)pkt)++ = htons(TFTP_ERROR);
		*((u16_t *)pkt)++ = htons(3);
		strcpy((char *)pkt, "File too large");
		pkt += 14 /*strlen("File too large")*/ + 1;
		len = pkt - xp;
		break;

	case STATE_BAD_MAGIC:
		xp = pkt;
		*((u16_t *)pkt)++ = htons(TFTP_ERROR);
		*((u16_t *)pkt)++ = htons(2);
		strcpy((char *)pkt, "File has bad magic");
		pkt += 18 /*strlen("File has bad magic")*/ + 1;
		len = pkt - xp;
		break;
         default:
                return;
	}

	NetSetEther(NetTxPacket, NetServerEther, PROT_IP);
	NetSetIP(NetTxPacket + ETHER_HDR_SIZE, NetServerIP,
					TftpServerPort, TftpOurPort, len);
	NetSendPacket(NetTxPacket, ETHER_HDR_SIZE + IP_HDR_SIZE + len);
}


static void
TftpHandler(u8_t * pkt, unsigned dest, unsigned src, unsigned len)
{
	if (dest != TftpOurPort)
		return;
	if (TftpState != STATE_RRQ && src != TftpServerPort)
		return;

	if (len < 2)
		return;
	len -= 2;
	switch (ntohs(*((u16_t *)pkt)++))
	{
	case TFTP_RRQ:
	case TFTP_WRQ:
	case TFTP_ACK:
	default:
		break;

	case TFTP_DATA:
		if (len < 2)
			return;
		len -= 2;
		TftpBlock = ntohs(*(u16_t *)pkt);

		if (TftpState == STATE_RRQ)
		{
			TftpState = STATE_DATA;
			TftpServerPort = src;
			TftpLastBlock = 0;

			if (TftpBlock != 1)	/* Assertion */
			{
				printf("\nTFTP error: First block is not "
						"block 1 (%d).\n", TftpBlock);
				printf("Starting again.\n\n");
				NetStartAgain();
				break;
			}
		}

		if (TftpBlock == TftpLastBlock)
		{
			/*
			 *	Same block again; ignore it.
			 */
			break;
		}

		TftpLastBlock = TftpBlock;
		TftpTimeoutCount = 0;

		/* ImageSize += len; */
		store_block(TftpBlock - 1, pkt + 2, len);

		/*
		 *	Acknoledge the block just received, which will prompt
		 *	the server for the next one.
		 */
		TftpSend();

		    if (TftpBlock % 8 == 0)
                        print_spinner(TftpBlock / 8);

		if (len < 512)
		{
			/*
			 *	We received the whole thing.  Try to
			 *	run it.
			 */
			printf("\nDone (%d kB/s)\n", 
                              (TftpBlock * get_hz()) / 
                                    (2 * (ticks_since_boot() - TftpStarted)));
			NetState = NETLOOP_SUCCESS;
		}
		break;

	case TFTP_ERROR:
		printf("\nTFTP error: '%s' (%d)\n",
					pkt + 2, ntohs(*(u16_t *)pkt));
		printf("Starting again.\n\n");
		NetStartAgain();
		break;
	}
}


static void
TftpTimeout(void)
{
	if (++TftpTimeoutCount >= TIMEOUT_COUNT)
	{
		printf("\nRetry count exceeded; starting again\n");
		NetStartAgain();
	}
	else
	{
		printf("t\b");
		NetSetTimeout(TIMEOUT * get_hz(), TftpTimeout);
		TftpSend();
	}
}


void
TftpStart(void)
{
#ifdef DEBUG
	printf(".\nServer ethernet address %02x:%02x:%02x:%02x:%02x:%02x\n",
			ServerEther[0],
			ServerEther[1],
			ServerEther[2],
			ServerEther[3],
			ServerEther[4],
			ServerEther[5]);
#endif /* DEBUG */

	printf("BOOTP/TFTP from server ");
	NetPrintIPaddr(NetServerIP);
	printf("; our IP address is ");
	NetPrintIPaddr(NetOurIP);
	printf(".\n");

	if (BootFile[0] == '\0')
	{
		int		i;

		for (i = 0; i < 4; i++)
		{
			static const char	hex[] = "0123456789ABCDEF";
			int		x;

			x = (NetOurIP >> (i * 8)) & 0xff;
			BootFile[i*2+0] = hex[x >> 4];
			BootFile[i*2+1] = hex[x & 0xf];
		}
		strcpy(&BootFile[8], ".img");

		printf("No file name; using '%s'.\n", BootFile);
	}
	else
	{
		printf("Filename '%s'.\n", BootFile);
	}

	printf("Loading: *\b");

#if 0
	{
		int		i;

		printf("Loading: >");
		for (i = 0; i < NDOTS; i++)
			printf(" ");
		printf("<\rLoading: >");
	}
#endif

	NetSetTimeout(TIMEOUT * get_hz(), TftpTimeout);
	NetSetHandler(TftpHandler);

	TftpServerPort = WELL_KNOWN_PORT;
	TftpTimeoutCount = 0;
	TftpState = STATE_RRQ;
	TftpOurPort = 1024 + (ticks_since_boot() % 3072);
        TftpStarted = ticks_since_boot();
	TftpSend();
}
