/* $Id: nvram.c,v 1.2 2000/11/15 17:44:55 apc Exp $ */
/* Copyright 2000  AG Electronics Ltd. */
/* This code is distributed without warranty under the GPL v2 (see COPYING) */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <console.h>
#include <ctype.h>

/* NVRAM layout
 * 
 * Environment variable record runs:
 * [length]NAME=value[length]NAME=value[0]\0
 * A deleted variable is:
 * [length]\0AME=value
 * 
 * When memory is full, we compact.
 * 
 */

static unsigned char *nvram_buffer;
static unsigned nvram_size;

int nvram_init (unsigned size)
{
    if (!nvram_buffer)
    {
	nvram_buffer = malloc (size);
	if (!nvram_buffer)
	    return -1;

	memset (nvram_buffer, 0, size);
	nvram_size = size;
    }
    return 0;
}

static char *find (unsigned char *base, const char *name, unsigned name_len)
{
    while (*base && (base - nvram_buffer < nvram_size))
    {
	unsigned length = *base;
	unsigned char *x;

	printf ("Length %d: ", length);
	if (base[1])
	{
	    for (x = base + 1; x < base + length; x++)
		if (isprint (*x))
		    console_outchar (*x);
	    console_outchar ('\n');
	}
	else
	    printf ("Empty\n");
	if (length > name_len && base[name_len + 1] == '=')
	{
	    if (strncmp (base + 1, name, name_len) == 0)
		break;
	}
	base += length;
    }
    return base;
}

int getenv (const char *name, char *buffer, unsigned size)
{
    unsigned char *base = nvram_buffer;
    unsigned name_len = strlen (name);

    printf ("Searching for \"%s\" (%d)\n", name, name_len);
    base = find (base, name, name_len);
    if (*base)
    {
	unsigned length = *base;

	if (length - name_len - 1 < size)
	    size = length - name_len - 1;
	memcpy (buffer, base + name_len + 2, size - 1);
	buffer[size - 1] = 0;
	printf ("Found \"%s\" (%d)", buffer, size - 1);
	return 0;
    }
    return -1;
}

static void nvram_merge (void)
{
    unsigned char *base = nvram_buffer;

    while (*base && (base - nvram_buffer < nvram_size))
    {
	unsigned length = base[0];
	unsigned char tmp = base[length];

	if (!base[1])
	    printf ("%d empty at %d\n", length, base - nvram_buffer);
	if (!base[1] && !base[length])
	{
	    base[0] = 0;
	    return;
	}
	else if (!base[1] && !base[length + 1] && (length + tmp < 255))
	{
	    printf ("Merging length %d at %d\n", length + tmp,
		    base - nvram_buffer);
	    base[0] = length + tmp;
	}
	else
	    base += length;
    }
}

static void nvram_delete (char *base)
{
    char *next = base + base[0];

    if (next[0])
    {
	base[1] = 0;
    }
    else
	/* Deleting the end entry */
	base[0] = 0;
    nvram_merge ();
}


void putenv (const char *name, const char *value)
{
    unsigned name_len = strlen (name);
    unsigned char *delete_me = NULL;

    /* Find an earlier setting, and delete it. This leaves a small window in
     * which a power failure can erase one nvram value. */
    delete_me = find (nvram_buffer, name, name_len);
    if (*delete_me)
	nvram_delete (delete_me);

    if (value)
    {
	unsigned char *base = nvram_buffer;
	unsigned value_len = strlen (value);
	unsigned slack = 8;
	unsigned slot_len = 2 + name_len + value_len;
	unsigned char *best_fit = NULL;

	/* Find a slot to take the variable. Allow 8 bytes slack */
	while (*base)
	{
	    unsigned length = *base;

	    if (base[1] == 0 && length >= slot_len
		    && length - slot_len < slack && length - slot_len != 1)
	    {
		slack = length - slot_len;
		best_fit = base;
		printf ("Slack of %d at %d\n", slack, base - nvram_buffer);
	    }
	    base += length;
	}
	if (best_fit)
	    base = best_fit;
	printf ("Slot at offset %d\n", base - nvram_buffer);
	memcpy (base + 1, name, name_len);
	base[1 + name_len] = '=';
	memcpy (base + 2 + name_len, value, value_len);
	if (!base[0])
	    base[0] = 2 + name_len + value_len;
	else
	{
	    printf ("Inserting length %d into slot %d, slack%d\n",
		    2 + name_len + value_len, base[0], slack);
	    base[2 + name_len + value_len] = slack;
	    base[3 + name_len + value_len] = 0;
	    base[0] = 2 + name_len + value_len;
	}

    }
}

void nvram_list (void)
{
    unsigned char *base = nvram_buffer;
    unsigned slack = 0;

    printf ("Environment variables:\n");
    while (*base && (base - nvram_buffer < nvram_size))
    {
	unsigned length = *base;

	if (base[1])
	{
	    unsigned char *x;

	    for (x = base + 1; x < base + length; x++)
		if (isprint (*x))
		    console_outchar (*x);
	    console_outchar ('\n');
	}
	else
	    slack += length;

	base += length;
    }
    printf ("Used %d/%d bytes, slack %d\n",
	    base - nvram_buffer, nvram_size, slack);
}
