/* $Id: pci.c,v 1.9 2000/11/15 17:44:55 apc Exp $ */
/* Copyright 2000  AG Electronics Ltd. */
/* This code is distributed without warranty under the GPL v2 (see COPYING) */

#include <stdio.h>
#include <pci.h>
#include <malloc.h>
#include <bsp.h>
#include <board.h>

/* In PCI Agent mode, we record the configuration we find, and the only change
   is to enable mastering */

#undef VERBOSE

#ifdef VERBOSE
#define PCI_VERBOSE(x) do{x;} while(0)
#else
#define PCI_VERBOSE(x) do{}while(0)
#endif

#define PCI_TYPE_MASK 0x7f
#define PCI_TYPE_DEVICE 0
#define PCI_TYPE_BRIDGE 1
#define PCI_ISBRIDGE(x) (((x) & PCI_TYPE_MASK) == PCI_TYPE_BRIDGE)
#define BRIDGE_IO_ALIGN 4096
#define BRIDGE_IO_ALIGN_UP(x) (((x) + BRIDGE_IO_ALIGN - 1) & ~(BRIDGE_IO_ALIGN-1))
#define BRIDGE_IO_ALIGN_DOWN(x) ((x) & ~(BRIDGE_IO_ALIGN-1))
#define BRIDGE_MM_ALIGN 1048576
#define BRIDGE_MM_ALIGN_UP(x) (((x) + BRIDGE_MM_ALIGN - 1) & ~(BRIDGE_MM_ALIGN-1))
#define BRIDGE_MM_ALIGN_DOWN(x) ((x) & ~(BRIDGE_MM_ALIGN-1))

int pci_default_latency = 0x80;
int pci_default_cacheline = 8;
struct pci_bus pci_root;

#ifndef PCI_AGENT
RODATA static const char *space_names[] = { "io", "mmio", "prefetch" };
static const unsigned space_alignments[] = { BRIDGE_IO_ALIGN,
    BRIDGE_MM_ALIGN, BRIDGE_MM_ALIGN
};
#endif

/* We create a pci_region structure for each region we must allocate. This includes
 * bridges.  These structures are freed when no longer needed. They are stored in
 * a sorted list in order of decreasing size. */
struct pci_region
{
    int bridge;			/* Set if the region is actually a bridge. This is required as
				 * a bridge device can also contain two regular regions. */
    int region;			/* The region number for a device. */
    u32 size;		        /* For a bridge, this is actually the greatest alignment
				 * required on the other side. */
    struct pci_device *device;	/* The containing device */
    struct pci_region *next;
};

static char last_bus_number;

#ifndef PCI_AGENT
/* Add a region into a queue, maintaining the sort order. */
static void insert_region (struct pci_region **head, int region,
			   unsigned size, int bridge,
			   struct pci_device *device)
{
    struct pci_region *data = malloc (sizeof (struct pci_region));

    if (!data)
    {
	printf ("Out of memory configuring PCI bus\n");
	bsp_indicate_dead ();
    }
    memset (data, 0, sizeof (struct pci_region));

    data->size = size;
    data->region = region;
    data->device = device;
    data->bridge = bridge;
    if (*head)
    {
	/* Insert at beginning */
	if ((*head)->size <= size)
	{
	    data->next = *head;
	    *head = data;
	    return;
	}
	else
	{
	    /* Look to see if it fits in the middle */
	    struct pci_region *p,
	    *q;

	    for (q = *head, p = (*head)->next; p; q = p, p = p->next)
	    {
		if (q->size >= size && p->size <= size)
		{
		    q->next = data;
		    data->next = p;
		    return;
		}
	    }
	    /* Fell off the end */
	    q->next = data;
	}
    }
    /* First entry */
    else
	*head = data;
}
#endif /* ! PCI_AGENT */

/* Create a bus object. The only trick is maintaining the numbering. */
static struct pci_bus *new_bus (struct pci_bus *parent)
{
    struct pci_bus *data = malloc (sizeof (struct pci_bus));

    if (!data)
    {
	printf ("Out of memory configuring PCI bus\n");
	bsp_indicate_dead ();
    }
    memset (data, 0, sizeof (struct pci_bus));

    data->parent = parent;
    data->next = parent->first_child;
    parent->first_child = data;
    data->bus = ++last_bus_number;
    data->have_io_space = parent->have_io_space;
    data->have_prefetch_space = 1;
    return data;
}

static struct pci_device *new_device (struct pci_bus *bus)
{
    struct pci_device *data = malloc (sizeof (struct pci_device));

    if (!data)
    {
	printf ("Out of memory configuring PCI bus\n");
	bsp_indicate_dead ();
    }
    memset (data, 0, sizeof (struct pci_device));

    data->bus = bus;
    data->next = bus->first_dev;
    bus->first_dev = data;

    return data;
}

static void configure_bus (struct pci_bus *parent,
			   struct pci_device *device)
{
    /* On bus 0, we don't configure device 0. Because it is us. */
    struct pci_bus *bus = new_bus (parent);
    int dev;
    u16 status;
#ifdef PCI_AGENT
    u8 busno;
#else
    u8 io_data;
    u16 prefetch_data;
#endif
    /* Configure bus numbers */
    bus->device = device;
    device->child = bus;

#ifdef PCI_AGENT
    pci_config_read_8(parent->bus, device->device, device->function, 0x19,
			&busno);
    bus->bus = busno;
#else
    /* As we don't know how many buses will be behind this one, we set the 
     * subordinate bus to 0xff. We can set the real value on the way back. */
    pci_config_write_8 (parent->bus, device->device, device->function, 0x19,
			bus->bus);
    pci_config_write_8 (parent->bus, device->device, device->function, 0x1a,
			0xff);
    pci_config_write_8 (parent->bus, device->device, device->function, 0x18,
			parent->bus);

    /* Check for existence of IO base, and disable */
    pci_config_write_8 (parent->bus, device->device, device->function, 0x1c,
			0xff);
    pci_config_write_8 (parent->bus, device->device, device->function, 0x1d,
			0x10);
    pci_config_read_8 (parent->bus, device->device, device->function, 0x1c,
		       &io_data);
    pci_config_write_32 (parent->bus, device->device, device->function,
			 0x30, 0);
    if (io_data == 0)
	bus->have_io_space = 0;
    /* And for prefetch space */
    pci_config_write_16 (parent->bus, device->device, device->function,
			 0x26, 0xf000);
    pci_config_write_16 (parent->bus, device->device, device->function,
			 0x24, 0xffff);
    pci_config_read_16 (parent->bus, device->device, device->function, 0x26,
			&prefetch_data);
    pci_config_write_32 (parent->bus, device->device, device->function,
			 0x28, 0);
    pci_config_write_32 (parent->bus, device->device, device->function,
			 0x2c, 0);
    if (prefetch_data == 0)
	bus->have_prefetch_space = 0;
    /* Put a default disabling value into memory space */
    pci_config_write_16 (parent->bus, device->device, device->function,
			 0x22, 0xf000);
    pci_config_write_16 (parent->bus, device->device, device->function,
			 0x20, 0xffff);
#endif /* ! PCI_AGENT */
    
    /* Check if the secondary is fast back to back capable */
    pci_config_read_16(parent->bus, device->device, device->function, 0x1e, &status);
    if (status & 0x80)
    {
        PCI_VERBOSE(printf("Setting fast back to back on bus %d\n", bus->bus));
        bus->fast_back_to_back = 1;   
    }

    for (dev = 0; dev < 32; dev++)
    {
	u32 data;
	int result = pci_config_read_32 (bus->bus, dev, 0, 0, &data);

	if (result == 0 && data != 0xffffffff)
	    pci_configure_device (bus, dev, 0);
    }

#ifndef PCI_AGENT
    /* At the end we know the real subordinate bus number */
    pci_config_write_8 (parent->bus, device->device, device->function, 0x1a,
			last_bus_number);

    /* And if all devices were fast back to back capable, set the flag */
    pci_config_write_16(parent->bus, device->device, device->function, 0x3e,
			    bus->fast_back_to_back ? 0x80 : 0);
    
    /* round sizes to bus limits, and move them into parent sizes. A bus
     * has the alignment constraints of its strictest device, which is also
     * the first in the list. */
    if (bus->spaces[PCI_REGION_IO])
    {
	ioaddr align = BRIDGE_IO_ALIGN;

	if (bus->spaces[PCI_REGION_IO]->size > BRIDGE_IO_ALIGN)
	    align = bus->spaces[PCI_REGION_IO]->size;
	insert_region (&parent->spaces[PCI_REGION_IO], PCI_REGION_IO,
		       align, 1, device);
    }
    if (bus->spaces[PCI_REGION_MMIO])
    {
	ioaddr align = BRIDGE_MM_ALIGN;

	if (bus->spaces[PCI_REGION_MMIO]->size > BRIDGE_MM_ALIGN)
	    align = bus->spaces[PCI_REGION_MMIO]->size;
	insert_region (&parent->spaces[PCI_REGION_MMIO], PCI_REGION_MMIO,
		       align, 1, device);
    }
    if (bus->spaces[PCI_REGION_PREFETCH])
    {
	ioaddr align = BRIDGE_MM_ALIGN;

	if (bus->spaces[PCI_REGION_PREFETCH]->size > BRIDGE_MM_ALIGN)
	    align = bus->spaces[PCI_REGION_PREFETCH]->size;
	if (parent->have_prefetch_space)
	    insert_region (&parent->spaces[PCI_REGION_PREFETCH],
			   PCI_REGION_PREFETCH, align, 1, device);
	else
	    insert_region (&parent->spaces[PCI_REGION_MMIO],
			   PCI_REGION_PREFETCH, align, 1, device);
    }
#endif /* ! PCI_AGENT */
}

void pci_configure_device (struct pci_bus *bus, int device, int function)
{
    struct pci_device *newdev = new_device (bus);
    int i;
    int no_regions = 0;
#ifndef PCI_AGENT
    u16 status;
#endif
    
    PCI_VERBOSE (printf ("Configuring pci bus %d device %d(%d)\n",
			 bus->bus, device, function));
    newdev->device = device;
    newdev->function = function;

#ifndef PCI_AGENT
    /* Set default values */
    pci_config_write_8 (bus->bus, device, function, 0xc,
			pci_default_cacheline);
    pci_config_write_8 (bus->bus, device, function, 0xd,
			pci_default_latency);

    /* Check for fast back to back */
    pci_config_read_16(bus->bus, device, function, 6, &status);
    if (! (status & 0x80))
    {
        PCI_VERBOSE(printf("Clearing FB2B\n"));
        bus->fast_back_to_back = 0;
    }
#endif /* ! PCI_AGENT */
        
    pci_config_read_8 (bus->bus, device, function, 0xe, &newdev->type);
    if ((newdev->type & PCI_TYPE_MASK) == PCI_TYPE_DEVICE)
	no_regions = 6;
    else if (PCI_ISBRIDGE (newdev->type))
	no_regions = 2;

    for (i = 0; i < no_regions; i++)
    {
#ifdef PCI_AGENT
	pci_config_read_32 (bus->bus, device, function, 0x10 + 4 * i,
			    &newdev->region[i]);
        if (newdev->region[i])
        {
            PCI_VERBOSE(printf ("[%d]: 0x%08lx\n", i, newdev->region[i]));
            if (newdev->region[i] & 1)
                newdev->region[i] &= ~3;
            else
                newdev->region[i] &= ~15;       
        }
#else
	ioaddr size;
        
#ifdef HAVE_PCI_ROM
        if ((bus->bus == PCI_ROM_BUS) && (device == PCI_ROM_DEVICE) &&
                (function == PCI_ROM_FUNCTION) && (i == PCI_ROM_REGION) &&
                RUNNING_FROM_PCI_ROM)
        {
            printf("Skipping PCI ROM region %02x:%02x(%x) %d\n",
                    bus->bus, device, function, i);
            continue;
        }
#endif
	pci_config_write_32 (bus->bus, device, function, 0x10 + 4 * i,
			     0x7fffffff);
	pci_config_read_32 (bus->bus, device, function, 0x10 + 4 * i,
			    &newdev->region[i]);
	if (newdev->region[i])
	{
            newdev->region[i] |= 0x80000000;
	    PCI_VERBOSE (printf ("[%d]: 0x%08lx - ", i, newdev->region[i]));

	    if (newdev->region[i] & 1)
		size = 1 + (3 | ~newdev->region[i]);
	    else
		size = 1 + (15 | ~newdev->region[i]);

	    PCI_VERBOSE (printf ("%ld bytes ", size));
	    if (newdev->region[i] & 1)
	    {
		PCI_VERBOSE (printf ("io space\n"));
		if (!bus->have_io_space)
		    printf
			("Warning %02x:%02x(%x): bus does not support io space\n",
			 bus->bus, device, function);
		else
		    insert_region (&bus->spaces[PCI_REGION_IO], i, size, 0,
				   newdev);
	    }
	    else
	    {
		PCI_VERBOSE (if (newdev->region[i] & 8)
			     printf ("prefetchable ");
			     printf ("memory space");
			     if ((newdev->region[i] & 6) == 2)
			     printf (", below 1MB"); printf ("\n"));

		if ((newdev->region[i] & 6) == 2)
		    printf
			("Warning %02x:%02x(%x): cannot configure mmio below 1MB\n",
			 bus->bus, device, function);
		else if ((newdev->region[i] & 8)
			 && bus->have_prefetch_space)
		    insert_region (&bus->spaces[PCI_REGION_PREFETCH], i,
				   size, 0, newdev);
		else
		    insert_region (&bus->spaces[PCI_REGION_MMIO], i, size,
				   0, newdev);
	    }
	}
#endif /* ! PCI_AGENT */
    }

    if (PCI_ISBRIDGE (newdev->type))
	configure_bus (bus, newdev);

    /* Multi-function device */
    if ((newdev->type & 0x80) && !function)
    {
	u32 data;

	for (i = 1; i < 8; i++)
	    if ((0 == pci_config_read_32 (bus->bus, device, i, 0, &data)) &&
		    (data != 0xffffffff))
		pci_configure_device (bus, device, i);
    }
}

#ifndef PCI_AGENT
/* Find the largest region that exactly fits the available space */
static struct pci_region *best_fit (struct pci_region **head, unsigned limit)
{
    struct pci_region *region;
    struct pci_region *prev;

    if (((limit & ~((*head)->size - 1)) == limit) || !(*head)->next)
    {
	region = *head;
	*head = region->next;
	return region;
    }

    for (prev = *head, region = prev->next; region && region->next;
	    prev = region, region = region->next)
    {
	if ((limit & ~(region->size - 1)) == limit)
	{
	    /* Exact fit. */
	    prev->next = region->next;
	    return region;
	}
    }

    prev->next = NULL;
    return region;
}

unsigned pci_assign_space (struct pci_bus *bus, int regtype, ioaddr limit,
		       ioaddr lower)
{
#ifdef VERBOSE
    unsigned wasted = 0;
#endif
    const char *name = space_names[regtype];

    while (bus->spaces[regtype])
    {
	struct pci_region *region;

	region = best_fit (&bus->spaces[regtype], limit);
	if (region->bridge)
	{
	    ioaddr upper = limit & ~(space_alignments[regtype] - 1);

	    PCI_VERBOSE(wasted += limit - upper);
	    upper--;
	    limit =
		pci_assign_space (region->device->child, region->region, upper,
			      lower);
	    PCI_VERBOSE (printf ("Bus %d %s space 0x%lx -> 0x%lx\n",
				 region->device->child->bus,
				 space_names[region->region], limit,
				 upper));
	    switch (region->region)
	    {
	    case PCI_REGION_IO:
		pci_config_write_8 (bus->bus, region->device->device,
				    region->device->function, 0x1c,
				    limit >> 8);
		pci_config_write_8 (bus->bus, region->device->device,
				    region->device->function, 0x1d,
				    upper >> 8);
		pci_config_write_32 (bus->bus, region->device->device,
				     region->device->function, 0x30, 0);
		break;
	    case PCI_REGION_MMIO:
		pci_config_write_16 (bus->bus, region->device->device,
				     region->device->function, 0x20,
				     limit >> 16);
		pci_config_write_16 (bus->bus, region->device->device,
				     region->device->function, 0x22,
				     upper >> 16);
		break;
	    case PCI_REGION_PREFETCH:
		pci_config_write_16 (bus->bus, region->device->device,
				     region->device->function, 0x24,
				     limit >> 16);
		pci_config_write_16 (bus->bus, region->device->device,
				     region->device->function, 0x26,
				     upper >> 16);
		pci_config_write_32 (bus->bus, region->device->device,
				     region->device->function, 0x28, 0);
		pci_config_write_32 (bus->bus, region->device->device,
				     region->device->function, 0x2c, 0);
		break;
	    }
	}
	else
	{
	    ioaddr upper = limit & ~(region->size - 1);

	    PCI_VERBOSE(wasted += limit - upper);
	    limit = upper - region->size;
	    upper--;
	    PCI_VERBOSE (printf
			 ("Assigning %d.%d(%d):%d %s space 0x%lx -> 0x%lx\n",
			  bus->bus, region->device->device,
			  region->device->function, region->region, name,
			  limit, upper));
	    pci_config_write_32 (bus->bus, region->device->device,
				 region->device->function,
				 0x10 + 4 * region->region, limit);
            /* Store for later */
            region->device->region[region->region] = limit;
	}
	if (limit < lower)
	    printf ("Error: %s went below lower (0x%08lx) on bus %d\n",
		    name, lower, bus->bus);
	free (region);
    }

    PCI_VERBOSE (wasted += limit & (space_alignments[regtype] - 1);
		 if (wasted)
		 printf ("Wasted 0x%x bytes %s space on bus %d\n",
			 wasted, name, bus->bus));
    return limit & ~(space_alignments[regtype] - 1);
}
#endif /* ! PCI_AGENT */

void pci_set_command(struct pci_bus *bus)
{
    struct pci_device *device;

    for (device = bus->first_dev; device; device = device->next)
    {
	pci_config_write_16 (bus->bus, device->device, device->function, 4,
			    bus->fast_back_to_back ? 0x029f :  0x009f);
        if (device->child)
            pci_set_command(device->child);
    }
}

RODATA static const char *class_00[] = {
    "Unknown", "VGA-compatible"
};
RODATA static const char *class_01[] = {
    "SCSI controller", "IDE controller", "Floppy controller",
    "IPI bus controller", "RAID controller"
};
RODATA static const char *class_02[] = {
    "Ethernet", "Token Ring", "FDDI", "ATM"
};
RODATA static const char *class_03[] = {
    "VGA", "XGA"
};
RODATA static const char *class_04[] = {
    "Video device", "Audio device"
};
RODATA static const char *class_05[] = {
    "RAM", "FLASH"
};
RODATA static const char *class_06[] = {
    "Host bridge", "ISA bridge", "EISA bridge", "MCA bridge",
    "PCI-to-PCI bridge", "PCMCIA bridge", "NuBus bridge", "CardBus bridge"
};
RODATA static const char *class_07[] = {
    "Serial controller", "Parallel port"
};
RODATA static const char *class_08[] = {
    "PIC", "DMA controller", "System timer", "RTC controller"
};
RODATA static const char *class_09[] = {
    "Keyboard controller", "Digitiser", "Mouse controller"
};
RODATA static const char *class_0c[] = {
    "IEEE 1394", "ACCESS.bus", "SSA", "USB controller",
    "Fibre Channel controller"
};

struct class_code
{
    const char **subclass;
    int no_subclass;
    const char *description;
};

const struct class_code classes[] = {
    {class_00, 2, "Unknown"},
    {class_01, 5, "Mass storage controller"},
    {class_02, 4, "Network controller"},
    {class_03, 2, "Display"},
    {class_04, 2, "Multimedia device"},
    {class_05, 2, "Memory controller"},
    {class_06, 8, "Bridge"},
    {class_07, 2, "Communications controller"},
    {class_08, 4, "System peripheral"},
    {class_09, 3, "Input device"},
    {NULL, 0, "Docking station"},
    {NULL, 0, "Processor"},
    {class_0c, 5, "Serial bus controller"}
};

void pci_print_bus (int depth, struct pci_bus *bus)
{
    struct pci_device *device;
    int i;

    for (device = bus->first_dev; device; device = device->next)
    {
	u16 vendor;
	u16 devno;
	u8 class;
	u8 subclass;
	u8 progif;

	printf ("%02x:%02x(%x) ", bus->bus, device->device,
		device->function);
	pci_config_read_16 (bus->bus, device->device, device->function, 0,
			    &vendor);
	pci_config_read_16 (bus->bus, device->device, device->function, 2,
			    &devno);
	printf ("0x%04x:0x%04x ", vendor, devno);
	pci_config_read_8 (bus->bus, device->device, device->function, 0xb,
			   &class);
	pci_config_read_8 (bus->bus, device->device, device->function, 0xa,
			   &subclass);
	pci_config_read_8 (bus->bus, device->device, device->function, 0x9,
			   &progif);
	for (i = 0; i < depth; i++)
	    printf ("  ");
	if (class < 0x0d)
	{
	    if (subclass < classes[class].no_subclass)
		printf ("%s\n", classes[class].subclass[subclass]);
	    else
		printf ("%s\n", classes[class].description);
	}
	else
	    printf ("Unknown\n");
	if (device->child)
	    pci_print_bus (depth + 1, device->child);
        else
        {
            int reg;
            for(reg = 0; reg < 6; reg++)
            {
                if (device->region[reg])
                {
                    u32 regval;
                    printf("                       ");
	            for (i = 0; i < depth; i++)
	                printf ("  ");
	            pci_config_read_32 (bus->bus, device->device,
                             device->function, 0x10 + reg*4,
			    &regval);
                    if (regval & 1)
                        printf(" %d:  io 0x%04lx\n", reg, device->region[reg]);
                        else
                        printf(" %d: 0x%08lx\n", reg, device->region[reg]);
                }
            }
        }
    }
}

void pci_scan (int argc, char **argv, void *data)
{
    pci_print_bus (0, &pci_root);
}

static struct pci_device *find_dev(struct pci_bus *bus,
        unsigned search_vendor, unsigned search_devno, unsigned *number)
{
    struct pci_device *device;

    for (device = bus->first_dev; device; device = device->next)
    {
	u16 vendor;
	u16 devno;

	pci_config_read_16 (bus->bus, device->device, device->function, 0,
			    &vendor);
	pci_config_read_16 (bus->bus, device->device, device->function, 2,
			    &devno);
        
        if (devno == search_devno && vendor == search_vendor)
        {
            if (*number)
                (*number)--;
            else
                return device;   
        }
        
        if (device->child)
        {
            struct pci_device *result =  find_dev(device->child, search_vendor,
                    search_devno, number);
            if (result)
                return result;
        }
    }
    return NULL;
}

struct pci_device *pci_find_device(unsigned vendor, unsigned device, unsigned number)
{
    return find_dev(&pci_root, vendor, device, &number);
}

